#! /usr/bin/python

"""
The Petmail agent runs on the local system, holds the private key, and acts
as an SMTP/POP/IMAP proxy for a local legacy MUA.
"""

import os.path

from twisted.application import service
from twisted.internet import defer, reactor
from twisted.python.components import implements
from petmail import addressbook, owner, message, permission
from petmail import interfaces, processors, crypto

class Agent(service.MultiService):
    """This is the main object which defines the Petmail agent. There will
    be exactly one of these in each agent process (it can accomodate
    multiple senders). This is the object which is serialized to record the
    persistent state of the service, including the contents of the
    addressbook and all the stored Permissions.

    """

    debugProcessing = False

    def __init__(self, basedir):
        service.MultiService.__init__(self)
        self.basedir = basedir
        self.me = None
        self.addressbook = None
        self.inboundProcessors = []
        self.outboundProcessors = []
        self.addressLookupAgents = []
        self.addressPublishAgents = []
        self.smtpScheme = "Mail:localhost:petmail-user@localhost"
        self.nntpScheme = None
        if 0:
            ds = debug.DebugService()
            ds.setServiceParent(self)

    def init(self):
        self.populateProcessors()
        self.addressbook = addressbook.AddressBook(self)
        self.crypto = crypto.EncryptionAgent(os.path.join(self.basedir,
                                                          "crypto"))
        self.me = owner.OwnerEntry()
        self.me.setServiceParent(self)
        self.me.setKey(self.crypto.generateKey("name"))
        self.me.genericPermission = permission.PermissionSlip()
        self.me.publishers = self.addressPublishAgents
        return self.me

    def setSMTP(self, scheme):
        self.smtpScheme = scheme

    def setAddressLookup(self, scheme):
        while self.addressLookupAgents:
            self.addressLookupAgents.pop()
        while self.addressPublishAgents:
            self.addressPublishAgents.pop()
        for s in scheme.split(","):
            la = addressbook.getLookupAgentFromString(s)
            self.registerAddressLookupAgent(la)
    def getAddressLookup(self):
        return ",".join([la.describe() for la in self.addressLookupAgents])

    def registerAddressLookupAgent(self, agent):
        self.addressLookupAgents.append(agent)
        self.addressPublishAgents.append(agent)
    def unregisterAddressLookupAgent(self, agent):
        self.addressLookupAgents.remove(agent)
        self.addressPublishAgents.remove(agent)

    def setNNTP(self, scheme):
        self.nntpScheme = scheme

    def getAddressBook(self, senderID):
        # return self.addressbooks[senderID]
        return self.addressbook
    def getSender(self, senderID):
        # return self.senders[senderID]
        return self.me

    def retrieveAddress(self, address):
        for method in self.addressLookupAgents:
            d = method.lookup(self, address)
            if isinstance(d, defer.Deferred):
                return d
            if d:
                return defer.succeed(d)
        else:
            raise interfaces.Unsendable, \
                  "none of my lookup agents could handle address '%s'" % \
                  address

    def populateProcessors(self):
        self.inboundProcessors = [pc() for pc in processors.allInbound]
        self.outboundProcessors = [pc() for pc in processors.allOutbound]

    def getInboundProcessors(self):
        return self.inboundProcessors[:]

    def getOutboundProcessors(self):
        return self.outboundProcessors[:]

    def getTicketAgents(self):
        # override this to provide ticket agents
        return []

    def acceptRFC822(self, rm, why):
        # override this to do something with incoming messages
        assert implements(rm, interfaces.IReceivedMessage)

    def addressBookChanged(self, addrbook):
        # override this to update something when the address book changes
        pass

    # sending

    def sendMessage(self, message):
        d = message.startProcessing()
        d.addCallback(self.deliverMessage)
        d.addErrback(self.sendMessageFailed, message)

    def deliverMessage(self, message):
        message.deliver()
        message.dump()

    def sendMessageFailed(self, why, message):
        print "sendMessageFailed", why

    # things that may happen during send

    def grantReturnPermission(self, entry, message):
        # make sure the user hasn't disabled this, then add Specific
        # Permission to this address book entry
        pass # TODO

    # receiving

    def receiveMessage(self, encrypted):
        """Feed an inbound message to the decryption/verification stage. If
        accepted, the message will be put on the inbox and
        notifyNewMessage() will be called.

        This will be called by a transport layer, triggered by a mailbox
        poll or SMTP receive from the outside world.

        @type encrypted: C{string}
        @param encrypted: the encrypted message blob
        """

        rm = message.ReceivedMessage(self, encrypted)
        senderID = rm.decrypt() # TODO: handle multiple owner keys
        rm.senderID = senderID
        # see if we already know the sender
        e = None
        if self.addressbook:
            e = self.addressbook.lookupID(senderID)
        if not e:
            # we don't know who they are. Create a temporary entry for them,
            # which will only be committed to our address book if we accept
            # an rfc822 message from them. 
            e = addressbook.AddressBookEntry(self.addressbook)
            e.ID = senderID
        rm.senderEntry = e

        d = rm.startProcessing()
        d.addCallback(self.receiveMessageDone)
        d.addErrback(self.receiveMessageFailed, rm)

    def receiveMessageDone(self, msg):
        """This is called when an inbound message has been fully processed.
        Everything interesting has already been done by the Processors, so
        all that remains is to log receipt (and possibly send back an
        acknowledgement)."""
        msg.dump()

    def receiveMessageFailed(self, why, msg):
        # an error occurred during processing, should log
        print "receiveMessageFailed", why
        msg.dump()

    def rejectRFC822(self, rm, why):
        # spam path, so this will only go into a log file, not bother the user
        print "rejectRFC822", why

    # things that may happen during receipt

    def addEntry(self, entry, messageThatCausedIt):
        self.addressbook.addEntry(entry)

    def grantInboundPermission(self, message):
        message.log(" not really")
        pass

    def run(self):
        self.startService()
        reactor.run()
        self.stopService()

    def stop(self):
        reactor.stop()

def run():
    a = Agent()
    a.startService()
    reactor.run()
    a.stopService()

if __name__ == '__main__':
    run()
