#! /usr/bin/python

from petmail import util
from base64 import encodestring, decodestring
from StringIO import StringIO

useBase64 = True

class UndecryptableError(Exception):
    """The message was not encrypted to our private key."""
class BadSignatureError(Exception):
    """The message had an invalid signature. Note that this does not mean
    the message was signed by an unknown public key. Such signatures cannot
    be verified, but are identified by returning None as a sigID rather than
    raising a BadSignatureError."""

# for this fake crypto, the keys and keyIDs made in triples:
#  "pubkey1", "privkey1", "keyid1"
# this lets us convert from keyids to keys without actually having a database
# of key values

def assertKeyID(s):
    assert(s.startswith("keyid"))

class PublicKey(object):
    def __init__(self, agent, keyid):
        if not agent.fake:
            assert len(keyid) == 40
        self.agent = agent
        self.keyid = keyid
    def getID(self):
        return self.keyid
    def getKey(self):
        return self.agent.getKey(self.keyid)
    def encryptAndSign(self, signer, plaintext):
        return self.agent.encryptAndSign(self, signer, plaintext)
    def __repr__(self):
        return "<PublicKey %s>" % self.getID()

class PrivateKey(PublicKey):
    def sign(self, plaintext):
        return self.agent.sign(self, plaintext)
    def pubkey(self):
        return PublicKey(self.agent, self.keyid)
    def __repr__(self):
        return "<PrivateKey %s>" % self.getID()

# verify must be done directly with the agent, as we don't know who it is
# encrypted to

class FakeEncryptionAgent(object):
    """I handle encryption needs for a PETmail agent. I wrap a GnuPG home
    directory and (and keyring) and provides methods to add keys, encrypt,
    decrypt, sign, and verify messages.

    My API is in terms of keyids.
    """

    fake = True

    def __init__(self, basedir=None):
        self.pubkeys = {}
        self.privkeys = {}

    def createKeyring(self):
        """Create my basedir and populate it with the standard GnuPG files.
        This must be done once before the encryption agent is useable.
        """
        pass
    
    def generateKey(self, name):
        key = util.generateMessageID()
        keyid = "keyid" + key
        self.privkeys[keyid] = "privkey" + key
        self.pubkeys[keyid] = "pubkey" + key
        return PrivateKey(self, keyid)

#    def getID(self, pubkey):
#        assert(pubkey.startswith("pubkey"))
#        return "keyid" + pubkey[6:]

    def addPubkey(self, pubkey):
        """I take a string with an armored public key (such as would come
        out of getKey(), below), extract the key, and add it to my keyring.
        After this, the key can be used as a target for encryption.
        """
        # do a gpg --import
        assert pubkey.startswith("pubkey")
        keyid = "keyid" + pubkey[6:]
        self.pubkeys[keyid] = pubkey
        return PublicKey(self, keyid)

    def getKey(self, keyid):
        """I return a string with an armored version of the given public
        key, suitable for inclusion in a message. This string has enough
        information to allow anyone who receives it to encrypt messages to
        this key's owner.
        """
        return self.pubkeys[keyid]

    def encryptAndSign(self, recipient, me, plaintext):
        """Encrypt and sign the message.

        @type recipientkey: C{string}
        @param recipientkey: the keyid to which the message should be encrypted
        @type ourkey: C{string}
        @param ourkey: the keyid that should be used to sign the message
        @type plaintext: C{string}
        @param plaintext: the (binary) message to be encrypted

        @rtype: C{string}
        @returns: a (binary) encrypted message
        """

        assert isinstance(me, PrivateKey)
        assert isinstance(recipient, PublicKey)
    
        if isinstance(plaintext, StringIO):
            plaintext = plaintext.read()
    
        s = "[%s,%s,%s]" % (recipient.getID(), me.getID(), plaintext)
        if useBase64:
            s = encodestring(s)
        return s


    def decrypt(self, crypttext):
        if isinstance(crypttext, StringIO):
            crypttext = crypttext.read()
        if useBase64:
            crypttext = decodestring(crypttext)
        assert(crypttext[0] == '[')
        assert(crypttext[-1] == ']')
        comma1 = crypttext.index(',')
        recipID = crypttext[1:comma1]
        comma2 = crypttext.index(',', comma1+1)
        sigID = crypttext[comma1+1:comma2]
        plaintext = crypttext[comma2+1:-1]
        if not self.privkeys.has_key(recipID):
            raise UndecryptableError, "message is not encrypted to us: recip=%s" % recipID
        return (PrivateKey(self, recipID), sigID, plaintext)

    def sign(self, me, plaintext):
        """Sign the message.

        @type ourkey: C{string}
        @param ourkey: the keyid that should be used to sign the message
        @type plaintext: C{string}
        @param plaintext: the (binary) message to be encrypted

        @rtype: C{string}
        @returns: a (binary) signed message
        """

        assert isinstance(me, PrivateKey)
        if isinstance(plaintext, StringIO):
            plaintext = plaintext.read()
        s = "[%s,%s]" % (me.getID(), plaintext)
        if useBase64:
            s = encodestring(s)
        return s

    def verify(self, crypttext):
        if isinstance(crypttext, StringIO):
            crypttext = crypttext.read()
        if useBase64:
            crypttext = decodestring(crypttext)
        assert(crypttext[0] == '[')
        assert(crypttext[-1] == ']')
        comma = crypttext.index(',')
        sigID = crypttext[1:comma]
        plaintext = crypttext[comma+1:-1]
        return (sigID, plaintext)

    def dumpSigned(self, crypttext):
        signer, plaintext = verify(crypttext)
        obj = util.decode(plaintext)

        print "signed by sigID:", signer.getID()
        print "obj:", obj
        if type(obj) == type({}) and obj.get("permission"):
            print "permission:", util.decode(obj['permission'])

import os, os.path, re
import GnuPGInterface

keygenInput = """Key-Type: DSA
Key-Length: 1024
# We are only testing syntax this time, so dry-run
Subkey-Type: ELG-E
Subkey-Length: 1024
Name-Real: %(name)s
Name-Comment: %(comment)s
Name-Email: %(email)s
"""
#Passphrase: %(passphrase)s
#%%dry-run

# for a first cut, this uses GnuPGInterface. It has problems: it blocks
# (os.waitpid), and doesn't actually do any of the useful --status-fd
# parsing that I need. It seems good enough to do encryption and decryption,
# but it doesn't give you enough information to do signature verification
# (except what you can parse yourself from stdout). Apart from being a
# frustratingly low-level interface, the options are too sticky (things like
# a recipients list should be set on a per-invocation basis, not on the
# whole object, unless the object is meant to be single-use), the read/write
# phases are distinct (which I think means large messages will cause
# deadlock), and it can't be used in a producer/consumer fashion (with
# implications for very large messages, albeit issues that are probably hard
# to fix anyway because of fundamental petmail design issues).

# This will be removed and replaced with a twisted ProcessProtocol-based
# implementation, that parses --status-fd to get reliable output. However
# that effort must wait until after I've enhanced twisted's spawnProcess
# framework to handle arbitrary additional file descriptors.

class GnuPGEncryptionAgent(object):
    fake = False
    def __init__(self, basedir):
        self.basedir = basedir
        # in general, we stick to 40-character fingerprints as IDs
        
        # .privkeys holds a list of our private key fingerprints
        self.privkeys = []

        self.createKeyring()

    def _getGPG(self):
        gpg = GnuPGInterface.GnuPG()
        gpg.options.homedir = self.basedir
        gpg.options.meta_interactive = 0
        gpg.options.extra_args.append("--no-secmem-warning")
        return gpg

    def createKeyring(self):
        if not os.path.exists(self.basedir):
            os.mkdir(self.basedir)
            os.chmod(self.basedir, 0700)
        # run gpg --list-keys, which will create pubring.gpg

    def generateKey(self, name):
        gpg = self._getGPG()
        proc = gpg.run(["--gen-key"],
                       create_fhs=['stdin', 'stderr', 'status',
                                   ])
        input = keygenInput % {'name': name,
                               'comment': "generated for PETmail",
                               'email': "petmail@thingy",
                               'passphrase': "",
                               }

        proc.handles['stdin'].write(input)
        proc.handles['stdin'].close()
        status = proc.handles['status'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        # status has a lot of "PROGRESS primegen + 0 0" lines
        # stderr has random things like "keyring foo.gpg created"

        for line in status.split("\n"):
            r = re.search(r'KEY_CREATED\s+(\w+)\s+(\w+)\b', line)
            if r:
                keyid = r.group(2)
                break
        else:
            raise RuntimeError("unable to create key:" + stderr)
        assert len(keyid) == 40
        self.privkeys.append(keyid)
        return PrivateKey(self, keyid)

    def addPubkey(self, pubkey):
        # feed key into stdin of --import
        # parse stderr for 'key XX: public key "name <addr>" imported'
        # then figure out keyid, probably with --list-colons on XX
        gpg = self._getGPG()
        proc = gpg.run(["--import"], create_fhs=['stdin', 'stderr', 'status'])
        proc.handles['stdin'].write(pubkey)
        proc.handles['stdin'].close()
        stderr = proc.handles['stderr'].read()
        status = proc.handles['status'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        lines = status.split("\n")
        for line in lines:
            importok = "[GNUPG:] IMPORT_OK "
            if line.startswith(importok):
                fields = line[len(importok):].split()
                reason, keyid = fields
                # could look at reason to check for newkey, unchanged, newsig
                break
        else:
            raise ValueError, "Unable to import key:" + stderr
        assert len(keyid) == 40
        return PublicKey(self, keyid)

    def _getLongID(self, otherid):
        gpg = self._getGPG()
        proc = gpg.run(["--with-colons", "--fixed-list-mode",
                        "--with-fingerprint", "--with-fingerprint",
                        "--list-keys", "0x%s" % otherid],
                       create_fhs=['stdout', 'stderr'])
        keys = proc.handles['stdout'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        longid = None
        for line in keys.split("\n"):
            fields = line.split(":")
            if fields[0] == "fpr":
                # the first 'fpr' line is the fingerprint
                longid = fields[9]
                assert len(longid) == 40
                return longid
        raise KeyError, "id '%s' not in keyring??: %s" % (otherid, stderr)

    def getKey(self, keyid):
        gpg = self._getGPG()
        proc = gpg.run(["--export", "0x%s" % keyid],
                       create_fhs=['stdout', 'stderr'])
        key = proc.handles['stdout'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        if not key:
            raise RuntimeError("gpg --export failed: " + stderr)
        return key

    def encryptAndSign(self, recipient, me, plaintext):
        assert isinstance(me, PrivateKey)
        assert isinstance(recipient, PublicKey)
        assert isinstance(plaintext, str)

        gpg = self._getGPG()
        gpg.options.recipients = [ "0x%s" % recipient.getID() ]
        gpg.passphrase = ""
        proc = gpg.run(['--local-user', "0x%s" % me.getID(),
                        '--sign', '--encrypt'],
                       create_fhs=['stdin', 'stdout', 'stderr'])
        proc.handles['stdin'].write(plaintext)
        proc.handles['stdin'].close()
        crypttext = proc.handles['stdout'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        if not crypttext:
            raise RuntimeError("gpg --sign --encrypt failed: " + stderr)
        return crypttext

    def decrypt(self, crypttext):
        """Decrypt the given message, returning a tuple of (recipient,
        signer, plaintext). May raise UndecryptableError if the message is
        not encrypted to one of our private keys.

        @type crypttext: C{string}
        @param crypttext: The (binary) encrypted message

        @returns: (recipient, sigID, plaintext)
                  recipient is a C{PrivateKey} to which the message was sent
                  sigID is a C{string} with the ID of the key that signed the
                   message
                  plaintext is a C{string} with the (binary) contents of the
                   message
        """

        assert isinstance(crypttext, str) # no StringIO yet
        gpg = self._getGPG()
        gpg.passphrase = ""
        proc = gpg.run(["--decrypt"],
                        create_fhs=['stdin', 'stdout', 'stderr', 'status'])
        proc.handles['stdin'].write(crypttext)
        proc.handles['stdin'].close()
        plaintext = proc.handles['stdout'].read()
        stderr = proc.handles['stderr'].read()
        status = proc.handles['status'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        for line in status.split("\n"):
            validsig = "[GNUPG:] VALIDSIG "
            if line.startswith(validsig):
                fields = line[len(validsig):].split()
                (fingerprint, sigdate, sigtime, expiretime, sigversion,
                 reserved, pubkeyalgo, hashalgo, sigclass,
                 primaryfingerprint) = fields
                signed = True
                sigid = primaryfingerprint
            errsig = "[GNUPG:] ERRSIG "
            if line.startswith(errsig):
                fields = line[len(errsig):].split()
                (keyid, pubkeyalgo, hashalgo, sigclass,
                 sigtime, rc) = fields
                signed = True
                sigid = None
            encto = "[GNUPG:] ENC_TO "
            if line.startswith(encto):
                fields = line[len(encto):].split()
                (keyid, keytype, keylength) = fields
                recipient = keyid
                # this keyid is a 16-byte id, not the full fingerprint
        if not signed:
            raise UndecryptableError("no valid signature found: " + stderr)
        if not recipient:
            raise UndecryptableError("message wasn't encrypted: " + stderr)
        recipID = self._getLongID(recipient)
        if sigid: assert len(sigid) == 40
        assert len(recipID) == 40
        return (PrivateKey(self, recipID), sigid, plaintext)


    def sign(self, me, plaintext):
        assert isinstance(me, PrivateKey)

        assert isinstance(plaintext, str) # no StringIO yet
        gpg = self._getGPG()
        gpg.passphrase = ""
        proc = gpg.run(['--local-user', "0x%s" % me.getID(),
                        '--sign'],
                       create_fhs=['stdin', 'stdout', 'stderr'])
        proc.handles['stdin'].write(plaintext)
        proc.handles['stdin'].close()
        crypttext = proc.handles['stdout'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        if not crypttext:
            raise RuntimeError("gpg --sign failed: " + stderr)
        return crypttext

    def verify(self, crypttext):
        """Verify the signature on the given message, returning a tuple of
        (sigID, plaintext). May raise BadSignatureError.

        @type crypttext: C{string}
        @param crypttext: The (binary) signed message

        @returns: (sigID, plaintext).

         sigID is a C{string} with the keyid of the private key used to sign
         the message, or None if the signature could not be checked because
         the key is not known.

         plaintext is a C{string} with the (binary) contents of the message.
        """

        assert isinstance(crypttext, str) # no StringIO yet
        gpg = self._getGPG()
        # we use --decrypt instead of --verify to get the plaintext out.
        # That text was not encrypted in the "crypttext", but it is
        # compressed and encoded in a way that would be silly to extract
        # with anything other than GPG
        proc = gpg.run(["--decrypt"],
                       create_fhs=['stdin', 'stdout', 'stderr', 'status'])
        proc.handles['stdin'].write(crypttext)
        proc.handles['stdin'].close()
        plaintext = proc.handles['stdout'].read()
        status = proc.handles['status'].read()
        stderr = proc.handles['stderr'].read()
        try:
            proc.wait()
        except IOError:
            print "GPG failed, stderr:", stderr
            raise

        for line in status.split("\n"):
            validsig = "[GNUPG:] VALIDSIG "
            if line.startswith(validsig):
                fields = line[len(validsig):].split()
                (fingerprint, sigdate, sigtime, expiretime, sigversion,
                 reserved, pubkeyalgo, hashalgo, sigclass,
                 primaryfingerprint) = fields
                signed = True
                sigid = primaryfingerprint
            errsig = "[GNUPG:] ERRSIG "
            if line.startswith(errsig):
                fields = line[len(errsig):].split()
                (keyid, pubkeyalgo, hashalgo, sigclass,
                 sigtime, rc) = fields
                signed = True
                sigid = None
        if not signed:
            raise BadSignatureError("no valid signature found:" + stderr)
        if sigid: assert len(sigid) == 40
        return (sigid, plaintext)

EncryptionAgent = GnuPGEncryptionAgent

if __name__ == '__main__':
    import sys
    agent = EncryptionAgent("/tmp/homedir")
    agent.createKeyring()
    if 0:
        data = open("/tmp/warner.key","r").read()
        pubkey = agent.addPubkey(data)
        print "KeyID:", pubkey
        key = pubkey.getKey()
        print "key data", len(key), "bytes"
    if 1:
        key = agent.generateKey("testkey")
        print key
        

