#! /usr/bin/python

from twisted.application import service
from twisted.python.components import implements

from petmail import crypto, transport, util, interfaces

class OwnerEntry(service.MultiService):
    """I hold information about one of the owner's identities. The agent
    holds a collection of these objects. Encoded copies will be published to
    the Name Servers.

    Every time you finish changing something about this record, you must
    call .publish() to push the changes out.

    All OwnerEntry objects are children of the Agent, so they can get at it
    with self.parent
    """

    __implements__ = (interfaces.IOwnerEntry,
                      service.MultiService.__implements__)

    def __init__(self):
        service.MultiService.__init__(self)
        self.myName = None # local pet name for this identity
        self.publishers = []
        self.inlets = []
        self.lastModified = util.now()
        self.privkey = None

        # fields which get published
        self.ID = None # derived from the pubkey
        self.name = "" # human names, affiliations: multiline text string
        self.addresses = [] # addresses which should resolve to this record
        # fields which are published in some indirect form
        self.genericPermission = None
        self.expirationTime = 120 # one month is more reasonable

    def modified(self):
        self.lastModified = util.now()

    def setKey(self, privkey):
        assert isinstance(privkey, crypto.PrivateKey)
        self.privkey = privkey
        self.ID = privkey.getID()
        self.modified()

    def changeKey(self):
        raise NotImplemented
        self.modified()

    def setName(self, name):
        self.name = name
        self.modified()

    def addAddress(self, addr):
        self.addresses.append(addr)
        self.modified()

    def publish(self):
        for p in self.publishers:
            for a in self.addresses:
                p.publish(a, self)

    def setGenericPermission(self, perm):
        self.genericPermission = perm
        self.modified()
    def setExpirationTime(self, t):
        self.expirationTime = t
        self.modified()

    def receiveMessage(self, encrypted):
        self.parent.receiveMessage(encrypted)

    def addInlet(self, inlet):
        if isinstance(inlet, str):
            inlet = transport.inletFromString(inlet, self.parent)
        assert implements(inlet, interfaces.IInlet)
        self.inlets.append(inlet)
        inlet.setServiceParent(self)
        self.modified()

    def removeInlet(self, inlet):
        if isinstance(inlet, str):
            for i in self.inlets:
                if i.string == inlet:
                    inlet = i
                    break
            else:
                raise KeyError("no such inlet '%s'" % inlet)
        self.inlets.remove(inlet)
        inlet.disownServiceParent()
        self.modified()

    def setInlet(self, s):
        """This presumes exactly one inlet."""
        if self.inlets:
            if self.inlets[0].string == s:
                return
            self.removeInlet(self.inlets[0])
        self.addInlet(s)
        self.modified()

    def encode(self):
        d = {'ID': self.ID,
             'pubkey': self.privkey.getKey(),
             'name': self.name,
             'addresses': self.addresses,
             'permission': None,
             'expires': util.now() + self.expirationTime,
             'transports': [inlet.encode() for inlet in self.inlets],
             }
        if self.genericPermission:
            d['permission'] = self.genericPermission.signed(self.privkey)
        return util.encode(d)

    def signed(self):
        """Return a signed string ready for uploading to a server or
        including in a a message. This is the serialized form of the
        IDRecord, and can be passed to addressbook.IDRecord to create such
        an object."""
        return self.privkey.sign(self.encode())

