#! /usr/bin/python

import cPickle
from StringIO import StringIO
from twisted.python.components import implements

from petmail import util, message, interfaces, ticket

class PermissionSlip(object):

    """Both sender and recipient have a copy of this PermissionSlip, with
    nearly identical contents.

    A newly-created PermissionSlip (using __init__) gives blanket permission
    to everybody.
    """

    debug = 0

    def __init__(self):
        # this slip allows .senderID to send mail to .receiverID

        # note that these are implicit: if the .ourPermission or
        # .theirPermission attributes in an AddressBookEntry point to this,
        # then senderID/receiverID are the owner of the AddressBook and the
        # target of the AddressBookEntry, or vice versa.

        self.senderID = None
        self.receiverID = None
        self.lastModified = util.now()

        # message attributes of the permission
        #self.maxrate = None # unlimited
        self.minIMG = 0 # inter-message-gap (seconds), 0 means unlimited.
        self.maxsize = None # unlimited
        self.contentTypes = ["*/*"]
        self.ticketURLs = [] # only populated in sender
        self.ticketTakers = {} # only populated in recipient

        # other attributes
        self.canGrantPermission = 1 # 3rd-party permission
        self.pedigree = None # where did this permission come from?

        # history, used to enforce maxrate
        self.lastSent = None # time last message was sent/received

    def modified(self):
        self.lastModified = util.now()

    def __repr__(self):
        if hasattr(self, 'ticketURLs'):
            url_string = ",".join(self.ticketURLs)
            url_string += "[%d]" % len(self.ticketURLs)
        else:
            url_string = self.getTicketTakersString()
        return "<PermissionSlip msize=%s URLs=%s>" % (self.maxsize,
                                                      url_string)

    def update(self, signed, cagent):
        signer, plaintext = cagent.verify(signed)
        # TODO: something?
        self.newFromEncoded(plaintext)

    def newFromEncoded(self, encoded):
        if not encoded:
            return
        d = util.decode(encoded)
        self.minIMG = d['IMG']
        self.maxsize = d['maxsize']
        self.contentTypes = d['contentTypes']
        self.ticketURLs = d['ticketVendors']
        self.canGrantPermission = d['canGrantPermission']
        self.modified()

    def setMinIMG(self, img):
        if img == self.minIMG:
            return
        self.minIMG = img
        self.modified()
    def setMaxSize(self, maxsize):
        if maxsize == self.maxsize:
            return
        self.maxsize = maxsize
        self.modified()

    def signed(self, us):
        """Return a string which encodes us"""
        d = {
            'IMG': self.minIMG,
            'maxsize': self.maxsize,
            'contentTypes': self.contentTypes,
            'ticketVendors': self.ticketTakers.keys(),
            'canGrantPermission': self.canGrantPermission,
            }
        return us.sign(util.encode(d))

    def getTicketTakersString(self):
        urls = self.ticketTakers.keys()
        urls.sort()
        return ",".join(urls)

    def setTicketTakersFromString(self, agent, str):
        new = []
        str = str.strip()
        for url in str.split(","):
            url.strip()
            if not url:
                continue
            if self.debug:
                print "setTicketTakersFromString: iter: '%s'" % url
            new.append(url)
            if url not in self.ticketTakers.keys():
                tt = ticket.TicketTaker(url, agent)
                self.addTicketTaker(tt)
        for url in self.ticketTakers.keys():
            if not url in new:
                self.removeTicketTaker(url)

    def addTicketTaker(self, taker):
        """The recipient calls this to add a new TicketTaker object to this
        slip. The sender who has access to this Permission Slip is thus
        allowed to earn entry to the mailbox by providing a ticket from the
        named URL."""
        self.ticketTakers[taker.url] = taker
        if self.debug:
            print "adding", taker
        self.modified()

    def removeTicketTaker(self, url):
        if self.debug:
            print "Removing", self.ticketTakers[url]
        del(self.ticketTakers[url])
        self.modified()

    def checkSizeAndIMG(self, msg):
        node = msg.getNode('rfc822')
        size = len(node['body'])
        contentType = "??"
        if self.maxsize and size > self.maxsize:
            why = "message is too large (%d > %d)" % (size, self.maxsize)
            raise message.NotPermitted, why
        # TODO: check self.contentTypes
        if self.minIMG and self.lastSent:
            lapsed = msg.when - self.lastSent
            if self.debug:
                print " IMG", self.minIMG
                print " lapsed", lapsed
            if lapsed < self.minIMG:
                raise message.NotPermitted, \
                      "last message was sent too recently"

    def checkOutboundMessage(self, msg):
        """Test an outbound message against these permissions. Raise an
        message.NotPermitted exception if the message cannot go through at
        all (it is too big, or too many have been sent recently). If tickets
        are needed, add the necessary information to the message's
        .requirements list.
        """
        assert implements(msg, interfaces.IMessage)
        requirements = []
        self.checkSizeAndIMG(msg)
        if self.debug:
            print "checkOutboundMessage:", self.ticketURLs
        if self.ticketURLs:
            # note that this message will need a ticket
            note = ('ticket', self.ticketURLs)
            requirements.append(note)
        return requirements

    def sendOutboundMessage(self, msg):
        """This method should be invoked when a message is actually sent. It
        will update the send history.
        """
        self.lastSent = msg.when


    def checkInboundMessage(self, msg):
        """Test an inbound message against our permissions. Raise an
        message.NotPermitted exception if the message is rejected for any
        reason. Verify that any necessary tickets are present.
        """
        assert implements(msg, interfaces.IReceivedMessage)
        node = msg.getNode('rfc822')
        self.checkSizeAndIMG(msg)

        why = "no ticket required"
        if self.ticketTakers:
            validTicket = 0
            tickets = ticket.extractTickets(node['tickets'],
                                            msg.agent.crypto)
            if self.debug:
                print "checkInboundMessage: tickets:", tickets
            for (url, sigID, t) in tickets:
                try:
                    taker = self.ticketTakers.get(url)
                    if not taker:
                        why = "The ticket was issued by an unknown " + \
                              "vendor (%s)" % url
                        raise ticket.InvalidTicket, why
                    # checkTicket raises InvalidTicket if it isn't valid
                    taker.checkTicket(sigID, t)
                    # ticket is valid
                    validTicket = 1
                    break
                except ticket.InvalidTicket, why:
                    print "ticket was invalid:", why.args[0]
            if not validTicket:
                raise message.NotPermitted, "ticket required but not supplied"
            why = "valid ticket from %s" % url
        # message is ok
        return why

    def acceptInboundMessage(self, msg):
        """This method should be invoked when a message is actually accepted.
        It will update the send history. Note that we should not track
        history on generic permissions, only on sender-specific ones.
        """
        self.lastSent = msg.when

    

def fromXML(source):
    """Parse XML into a new PermissionSlip object"""
    raise NotImplemented

def fromPickle(source):
    """Unpickle a PermissionSlip object"""
    slip = cPickle.loads(source)
    return slip

def decode(source):
    return fromPickle(source)
