#! /usr/bin/python

"""
This file contains various forms of Transport objects, which are used to get
an encapsulated message body (a binary string) to their receiver. The most
likely choice is a MIME-encoded rfc822 message send via the legacy SMTP
infrastructure, but anything which reliably gets the blob to the other end
will work.

The Transport is also responsible for doing an end-to-end acknowledgement,
retransmitting if necessary.
"""

import os, os.path
from email import Encoders, Parser, Message

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from twisted.python import log
from twisted.spread import pb
from twisted.cred import portal
from twisted.application import service, internet
from twisted.internet import protocol, error

from twisted.protocols.smtp import sendmail
from twisted.mail import relaymanager

from petmail import pbutil, interfaces

# the recipient adds an Inlet to their agent
# the Inlet produces a transport string (.encode)
# the transport string is published in the ID Record
# the sender gets the ID Record and extracts the transport string
# the transport string has a prefix and the args
# the prefix is looked up in knownTransports and used to create a Transport

knownInlets = {}
knownTransports = {}

def registerInlet(prefix, factory):
    knownInlets[prefix] = factory
def registerTransport(prefix, factory):
    knownTransports[prefix] = factory

def inletFromString(string, agent):
    colon = string.index(":")
    type = string[:colon]
    args = string[colon+1:]
    inlet = knownInlets[type](args, agent)
    inlet.string = string
    return inlet

def transportFromString(string, agent):
    if not string:
        return Null(None, agent)
    colon = string.index(":")
    type = string[:colon]
    args = string[colon+1:]
    transport = knownTransports[type](args, agent)
    transport.string = string # TODO? why?
    return transport

class BaseInlet(service.MultiService):
    __implements__ = (interfaces.IInlet, service.MultiService.__implements__)

    def encode(self):
        return self.description

dummyInlets = {}
class DummyInlet(BaseInlet):
    """DummyInlet/DummyTransport are paired like PTY devices. Both live in
    the same process, and writing to one causes the message to be emitted on
    the other. Used only for testing.

    Dummy:name
    """

    def __init__(self, args, agent):
        BaseInlet.__init__(self)
        self.name = args
        dummyInlets[self.name] = self
        self.description = "Dummy:%s" % self.name
    def accept(self, encrypted):
        self.parent.receiveMessage(encrypted)

class DummyTransport:
    __implements__ = interfaces.ITransport,
    def __init__(self, args, agent):
        self.name = args
    def send(self, encapsulated):
        inlet = dummyInlets[self.name]
        inlet.accept(encapsulated)
registerInlet("Dummy", DummyInlet)
registerTransport("Dummy", DummyTransport)


class Null:
    __implements__ = interfaces.ITransport,

    def __init__(self, args, agent):
        pass
    def send(self, encapsulated):
        print "discarding message"
registerTransport("Null", Null)



class PBInlet(BaseInlet, pb.Root):
    __implements__ = (BaseInlet.__implements__, pb.Root.__implements__)

    def __init__(self, args, agent):
        BaseInlet.__init__(self)
        self.description = "PB:" + args
        hostname, portnum = args.split(":")
        server = pbutil.listen(self, int(portnum))
        server.setServiceParent(self)

    def remote_message(self, encrypted):
        self.parent.receiveMessage(encrypted)
        
    def startService(self):
        BaseInlet.startService(self)

    def stopService(self):
        BaseInlet.stopService(self)
        print "stopped listening"
        # probably need to wait a moment here if the socket is going to
        # be re-opened right away

registerInlet("PB", PBInlet)


class PBTransport:
    __implements__ = interfaces.ITransport,
    def __init__(self, args, agent):
        host, portnum = args.split(":")
        self.host = host
        self.portnum = int(portnum)
    def send(self, encapsulated):
        print "sending"
        d = pbutil.connect(self.host, self.portnum)
        d.addCallback(self._connected, encapsulated)
        d.addErrback(log.err)
    def _connected(self, ref, encapsulated):
        print "connected"
        d = ref.callRemote("message", encapsulated)
        d.addCallback(self._sent, ref)
        d.addErrback(log.err)
    def _sent(self, result, ref):
        print "sent"
        ref.broker.transport.loseConnection()
    def __repr__(self):
        return "petmail.transport.PB<%s:%d>" % (self.host, self.portnum)
registerTransport("PB", PBTransport)


class SMTPEncapsulatedInlet(BaseInlet):

    # the next three methods may raise an exception, in which case the error
    # should be logged and the message dropped (or perhaps put in a log
    # directory for a limited period of time)

    def gotText(self, text):
        """This method should be called when the inlet has received an
        incoming message ready to be parsed by the standard 'email' package..
        This particular function should be given the message as a string."""
        f = StringIO(text)
        self.gotFile(f)

    def gotFile(self, fp):
        """This function receives the message as an open file, 
        """
        parser = Parser.Parser()
        m = parser.parse(fp)
        self.gotMessage(m)

    def gotMessage(self, m):
        """This method is called with a complete Message object."""
        
        assert not m.is_multipart()
        assert m.get_content_type() == "application/x-petmail"
        payload = m.get_payload(decode=True)
        self.parent.receiveMessage(payload)

class MaildirInlet(SMTPEncapsulatedInlet):
    """I watch a maildir for inbound messages."""

    pollInterval = 10 # seconds
    startedTimer = False

    def __init__(self, args, agent):
        SMTPEncapsulatedInlet.__init__(self)
        d, address = args.split(":")
        self.newdir = os.path.join(os.path.expanduser(d), "new")
        self.olddir = os.path.join(os.path.expanduser(d), "cur")
        self.description = "SMTP:%s" % address

    def startService(self):
        if not self.startedTimer:
            timer = internet.TimerService(self.pollInterval, self.check)
            timer.setServiceParent(self)
            self.startedTimer = True
        SMTPEncapsulatedInlet.startService(self)

    def check(self):
        for f in os.listdir(self.newdir):
            newfile = os.path.join(self.newdir, f)
            oldfile = os.path.join(self.olddir, f)
            self.gotFile(open(newfile, "r"))
            os.rename(newfile, oldfile)
                      
registerInlet("Maildir", MaildirInlet)

class MailSocketProtocol(protocol.Protocol):
    def __init__(self):
        self.data = ""
    def dataReceived(self, data):
        self.data += data
    def connectionLost(self, reason):
        if reason.check(error.ConnectionDone):
            self.factory.parent.gotText(self.data)

class MailSocket(SMTPEncapsulatedInlet):
    # [MailSocket:]9990:alice@foo.com
    # socketnum : inletaddress

    def __init__(self, args, agent):
        SMTPEncapsulatedInlet.__init__(self)
        portnum, address = args.split(":")
        portnum = int(portnum)
        f = protocol.ServerFactory()
        f.protocol = MailSocketProtocol
        f.parent = self
        server = internet.TCPServer(portnum, f)
        server.setServiceParent(self)
        self.description = "SMTP:%s" % address
registerInlet("MailSocket", MailSocket)


# SMTP-encapsulating output. The Agent has a .smtpScheme string which defines
# a preference for how we should do outbound SMTP. This scheme is
# "Mail:relay:fromaddr" for simple mailing, and "Mixminion:??" for
# send-through-mixminion

smtpSchemes = {}
def makeSMTPTransport(args, agent):
    scheme, schemeArgs = agent.smtpScheme.split(":", 1)
    transport = smtpSchemes[scheme](args, agent)
    return transport
registerTransport("SMTP", makeSMTPTransport)

class SMTPEncapsulatingTransport(object):
    """This is a base class for Transports which encapsulate PETmail messages
    in SMTP payloads (rfc822-formatted strings) so they can be mailed to the
    recipient's Inlet. Subclasses are responsible for actually sending the
    SMTP message.
    """
    __implements__ = interfaces.ITransport,

    def encode(self, encapsulated):
        """This puts the PETmail message body into a MIME payload. It returns
        a Message object, so you probably want to do .as_string() to extract
        the text."""

        m = Message.Message()
        m.set_type("application/x-petmail")
        m.set_payload(encapsulated)
        Encoders.encode_base64(m)
        m['Subject'] = "PETmail Message"
        m['X-PETmail-URL'] = "http://www.lothar.com/tech/spam/index.html"
        return m

class JustMailIt(SMTPEncapsulatingTransport):
    """This subclass uses twisted.mail to deliver the message via SMTP to a
    designated relay host. For normally-abled unix boxes that run an MTA like
    sendmail, this defaults to 'localhost'. Others will need to set it to a
    host which *is* capable of sending mail.
    """

    def __init__(self, args, agent):
        scheme, self.smarthost, self.fromEmail = agent.smtpScheme.split(":")
        assert scheme == "Mail"
        self.dest = args

    def send(self, encapsulated):
        """This returns a Deferred which will fire when the message has been
        sent successfully."""

        m = self.encode(encapsulated)
        m['From'] = self.fromEmail
        m['To'] = self.dest
        s = m.as_string()
        return sendmail(self.smarthost, self.fromEmail, self.dest, s)
smtpSchemes["Mail"] = JustMailIt

class MailAnonymously(SMTPEncapsulatingTransport):
    """This subclass sends the message out through a remailer network, making
    it nigh-impossible to figure out where it came from. The only information
    available to the recipient (or an eavesdropper) is the contents of the
    IDRecord."""

    pass
smtpSchemes["Mixminion"] = MailAnonymously

class UsenetDropbox(SMTPEncapsulatingTransport):

    """This subclass posts the message to a Usenet group, probably
    alt.anonymous.messages, where the recipient can pick it up at their
    (fairly anonymous) leisure. The sender is not particularly anonymous
    unless they send the message through a remailer network."""

    pass

# TODO: the SMTP Tranports have two kinds of configurable state: that which
# comes from the sender's configuration, and that which comes from the
# recipient's TransportRecord.

# recipient defines:
#  "I want messages via SMTP at address bob@foo.com"
#  "I want messages to be posted to alt.anonymous.messages with subject
#   '123'"

# sender defines:
#  "All SMTP messages should be sent with a From: address of alice@foo.com"
#  "All SMTP messages should be sent through smarthost smtp.isp.com"
#  "All SMTP messages should be sent through mixminion"
#  "Usenet posts should be posted to nntp.isp.com"
#  "Usenet posts should be mail-to-newsed through mixminion and m2n.blah.com"

# Likewise, the Inlet contains state that is purely local, which should not
# be published in their TransportRecord. The published data is listed above.

# recipient-local data:
#  "I get SMTP mail by sucking POP from user/pass @pop.isp.com every 5min"
#  " or by IMAP"
#  "I get SMTP by listening on port 25"
#  "I get SMTP by sucking on a maildir at ~/Maildir-petmail every 5 minutes"
#  "I get SMTP by watching ~/Maildir-petmail for changes with FAM or such"
#  "I get SMTP by one of the above, and sometimes doing nymserver-IMAP
#    polling when the nymserver tells me there are a lot of messages queued,
#    and oh by the way please keep the nymserver supplied with SURBs"
#  "I get Usenet posts by polling nntp.isp.com every 30 minutes"
#  "I get SMTP by listening on localhost:9990 or the ~/.petmail-inlet UNIX
#    socket, because my .forward/.qmail says '|nc localhost 9990'"
